from pyramid_debugtoolbar.panels import DebugPanel

from pyramid.interfaces import ICSRFStoragePolicy
from ... import DualCookieCSRFStoragePolicy

_ = lambda x: x


class CSRFMultiSchemeDebugPanel(DebugPanel):
    """
    Sample debug panel
    """

    name = "CSRFMultiScheme"
    has_content = None
    template = "pyramid_csrf_multi_scheme.debugtoolbar.panels:templates/csrf_multi_scheme.dbtmako"

    # only query the policy once
    _policy = None

    # stash
    _cookie_names = None

    response = None  # we may never have a response

    def __init__(self, request):

        self.data = {
            "request_data": {
                "cookie_secure": {"in": None, "out": None, "cookie_name": None},
                "cookie_http": {"in": None, "out": None, "cookie_name": None},
            }
        }
        policy = request.registry.queryUtility(ICSRFStoragePolicy)
        if isinstance(policy, DualCookieCSRFStoragePolicy):
            self._policy = policy
            self._cookie_names = [policy.cookie_name_secure, policy.cookie_name_http]

            self.has_content = True

            self.data["request_data"]["cookie_secure"]["in"] = request.cookies.get(
                policy.cookie_name_secure
            )
            self.data["request_data"]["cookie_secure"][
                "cookie_name"
            ] = policy.cookie_name_secure

            self.data["request_data"]["cookie_http"]["in"] = request.cookies.get(
                policy.cookie_name_http
            )
            self.data["request_data"]["cookie_http"][
                "cookie_name"
            ] = policy.cookie_name_http

            # `process_response` fires before the set_cookie callbacks, so use this hack...
            def finished_callback(request):
                self._process_response_deferred()

            request.add_finished_callback(finished_callback)

    def process_response(self, response):
        """`
        process_response` fires before the set_cookie callbacks
        so we use a hack from pyramid_debugtoolbar's HeaderDebugPanel

        1. `__init__()` sets a callback to `_process_response_deferred()`
        2. `process_response()` stashes the response
        3. `_process_response_deferred` has the headers
        """
        self.response = response

    def _process_response_deferred(self):
        """
        `process_response` fires before the set_cookie callbacks
        so we use a hack from pyramid_debugtoolbar's HeaderDebugPanel
        """
        if self.response is not None:
            cookies = [i for i in self.response.headerlist if i[0] == "Set-Cookie"]
            # [('Set-Cookie', 'ci_ss_ck=eXrTEtumW4Jal3FyEXtkJijofCAyOnayzGQiPG7IAjukycfE; Domain=.dev.example.com; Path=/'),
            #  ('Set-Cookie', 'ci_s_ck=iG6dqE5mff_iR2Y95o5eENuVdC; Domain=.dev.example.com; Path=/')
            #  ]
            for ck in cookies:
                try:
                    ck_parts = ck[1].split(";")
                    (ck_name, ck_value) = ck_parts[0].split(
                        "=", 1
                    )  # maxsplit 1, because the value could be an encoding with == as a right pad
                    if ck_name in self._cookie_names:
                        if (
                            ck_name
                            == self.data["request_data"]["cookie_secure"]["cookie_name"]
                        ):
                            self.data["request_data"]["cookie_secure"]["out"] = ck_value
                        elif (
                            ck_name
                            == self.data["request_data"]["cookie_http"]["cookie_name"]
                        ):
                            self.data["request_data"]["cookie_http"]["out"] = ck_value
                except Exception as exc:
                    # TODO: logging
                    pass

    @property
    def nav_title(self):
        return _(self.name)

    @property
    def title(self):
        return _(self.name)

    @property
    def url(self):
        return ""

    def render_content(self, request):
        return DebugPanel.render_content(self, request)



# stdlib
import unittest

# pypi
from pyramid.config import Configurator
from pyramid.interfaces import ICSRFStoragePolicy

# our local tested package
from pyramid_csrf_multi_scheme import DualCookieCSRFStoragePolicy


# ==============================================================================


class DummyRequest(object):
    registry = None
    session = None
    response_callbacks = None
    scheme = None

    def __init__(self, registry=None, session=None, scheme="http"):
        self.registry = registry
        self.session = session
        self.cookies = {}
        self.scheme = scheme
        self.response_callbacks = []

    def add_response_callback(self, callback):
        self.response_callbacks.append(callback)

    def response_callback(self, request, response):
        while len(self.response_callbacks):
            callback = self.response_callbacks.pop()
            callback(request, response)


class MockResponse(object):
    def __init__(self):
        self.headerlist = []


class TestCookieCSRFStoragePolicy_InsecureHttp(unittest.TestCase):
    """tests on the insecure channel"""

    def _makeOne(self, **kw):
        return DualCookieCSRFStoragePolicy(**kw)

    def test_register_cookie_csrf_policy(self):
        config = Configurator()
        config.set_csrf_storage_policy(self._makeOne())
        config.commit()
        policy = config.registry.queryUtility(ICSRFStoragePolicy)
        self.assertTrue(isinstance(policy, DualCookieCSRFStoragePolicy))

    def test_get_cookie_csrf_with_no_existing_cookie_sets_cookies(self):
        response = MockResponse()
        request = DummyRequest()

        policy = self._makeOne()
        token = policy.get_csrf_token(request)
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [("Set-Cookie", "csrf_http={}; Path=/; SameSite=Lax".format(token))],
        )

    def test_get_cookie_csrf_nondefault_samesite(self):
        response = MockResponse()
        request = DummyRequest()

        policy = self._makeOne(samesite=None)
        token = policy.get_csrf_token(request)
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist, [("Set-Cookie", "csrf_http={}; Path=/".format(token))]
        )

    def test_existing_cookie_csrf_does_not_set_cookie(self):
        request = DummyRequest()
        request.cookies = {"csrf_http": "e6f325fee5974f3da4315a8ccf4513d2"}

        policy = self._makeOne()
        token = policy.get_csrf_token(request)

        self.assertEqual(token, "e6f325fee5974f3da4315a8ccf4513d2")
        self.assertEqual(len(request.response_callbacks), 0)

    def test_new_cookie_csrf_with_existing_cookie_sets_cookies(self):
        request = DummyRequest()
        request.cookies = {"csrf_http": "e6f325fee5974f3da4315a8ccf4513d2"}

        policy = self._makeOne()
        token = policy.new_csrf_token(request)

        response = MockResponse()
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [("Set-Cookie", "csrf_http={}; Path=/; SameSite=Lax".format(token))],
        )

    def test_get_csrf_token_returns_the_new_token(self):
        request = DummyRequest()
        request.cookies = {"csrf_http": "foo"}

        policy = self._makeOne()
        self.assertEqual(policy.get_csrf_token(request), "foo")

        token = policy.new_csrf_token(request)
        self.assertNotEqual(token, "foo")
        self.assertEqual(token, policy.get_csrf_token(request))

    def test_check_csrf_token(self):
        request = DummyRequest()

        policy = self._makeOne()
        self.assertFalse(policy.check_csrf_token(request, "foo"))

        request.cookies = {"csrf_http": "foo"}
        self.assertTrue(policy.check_csrf_token(request, "foo"))
        self.assertFalse(policy.check_csrf_token(request, "bar"))


class TestCookieCSRFStoragePolicy_SecureHTTPS(unittest.TestCase):
    """tests on the SECURE HTTPS"""

    def _makeOne(self, **kw):
        return DualCookieCSRFStoragePolicy(**kw)

    def test_register_cookie_csrf_policy(self):
        config = Configurator()
        config.set_csrf_storage_policy(self._makeOne())
        config.commit()
        policy = config.registry.queryUtility(ICSRFStoragePolicy)
        self.assertTrue(isinstance(policy, DualCookieCSRFStoragePolicy))

    def test_get_cookie_csrf_with_no_existing_cookie_sets_cookies(self):
        response = MockResponse()
        request = DummyRequest(scheme="https")

        policy = self._makeOne()
        token_secure = policy.get_csrf_token(request)
        token_http = policy.get_csrf_token_scheme(request, "http")
        request.response_callback(request, response)

        self.assertEqual(
            response.headerlist,
            [
                (
                    "Set-Cookie",
                    "csrf_https={}; Path=/; secure; SameSite=Lax".format(token_secure),
                ),
                ("Set-Cookie", "csrf_http={}; Path=/; SameSite=Lax".format(token_http)),
            ],
        )

    def test_get_cookie_csrf_nondefault_samesite(self):
        response = MockResponse()
        request = DummyRequest(scheme="https")

        policy = self._makeOne(samesite=None)
        token_secure = policy.get_csrf_token(request)
        token_http = policy.get_csrf_token_scheme(request, "http")
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [
                ("Set-Cookie", "csrf_https={}; Path=/; secure".format(token_secure)),
                ("Set-Cookie", "csrf_http={}; Path=/".format(token_http)),
            ],
        )

    def test_existing_cookie_csrf_does_not_set_cookie(self):
        request = DummyRequest(scheme="https")
        request.cookies = {"csrf_https": "e6f325fee5974f3da4315a8ccf4513d2"}

        policy = self._makeOne()
        token_secure = policy.get_csrf_token(request)

        self.assertEqual(token_secure, "e6f325fee5974f3da4315a8ccf4513d2")
        self.assertEqual(len(request.response_callbacks), 0)

    def test_new_cookie_csrf_with_existing_cookie_sets_cookies(self):
        request = DummyRequest(scheme="https")
        request.cookies = {"csrf_https": "e6f325fee5974f3da4315a8ccf4513d2"}

        policy = self._makeOne()
        token_secure = policy.new_csrf_token(request)
        token_http = policy.get_csrf_token_scheme(request, "http")

        response = MockResponse()
        request.response_callback(request, response)
        self.assertEqual(
            response.headerlist,
            [
                (
                    "Set-Cookie",
                    "csrf_https={}; Path=/; secure; SameSite=Lax".format(token_secure),
                ),
                ("Set-Cookie", "csrf_http={}; Path=/; SameSite=Lax".format(token_http)),
            ],
        )

    def test_get_csrf_token_returns_the_new_token(self):
        request = DummyRequest(scheme="https")
        request.cookies = {"csrf_https": "foo"}

        policy = self._makeOne()
        self.assertEqual(policy.get_csrf_token(request), "foo")

        token_secure = policy.new_csrf_token(request)
        self.assertNotEqual(token_secure, "foo")
        self.assertEqual(token_secure, policy.get_csrf_token(request))

    def test_check_csrf_token(self):
        request = DummyRequest(scheme="https")

        policy = self._makeOne()
        self.assertFalse(policy.check_csrf_token(request, "foo"))

        request.cookies = {"csrf_https": "foo"}
        self.assertTrue(policy.check_csrf_token(request, "foo"))
        self.assertFalse(policy.check_csrf_token(request, "bar"))



from __future__ import print_function
from __future__ import unicode_literals

# stdlib
import re
import unittest

# pyramid testing requirements
from pyramid import testing
from pyramid.exceptions import ConfigurationError
from pyramid.response import Response
from pyramid.request import Request


# ------------------------------------------------------------------------------


# used to ensure the toolbar link is injected into requests
re_toolbar_link = re.compile(r'(?:href="http://localhost)(/_debug_toolbar/[\d]+)"')


class TestDebugtoolbarPanel(unittest.TestCase):
    def setUp(self):
        self.config = config = testing.setUp()
        config.add_settings(
            {"debugtoolbar.includes": ["pyramid_csrf_multi_scheme.debugtoolbar"]}
        )
        config.include("pyramid_debugtoolbar")
        self.settings = config.registry.settings

        # create a view
        def empty_view(request):
            return Response(
                "<html><head></head><body>OK</body></html>", content_type="text/html"
            )

        config.add_view(empty_view)

    def tearDown(self):
        testing.tearDown()

    def test_panel_injected(self):

        # make the app
        app = self.config.make_wsgi_app()
        # make a request
        req1 = Request.blank("/")
        req1.remote_addr = "127.0.0.1"
        resp1 = req1.get_response(app)
        self.assertEqual(resp1.status_code, 200)
        self.assertIn("http://localhost/_debug_toolbar/", resp1.text)

        # check the toolbar
        links = re_toolbar_link.findall(resp1.text)
        self.assertIsNotNone(links)
        self.assertIsInstance(links, list)
        self.assertEqual(len(links), 1)
        toolbar_link = links[0]

        req2 = Request.blank(toolbar_link)
        req2.remote_addr = "127.0.0.1"
        resp2 = req2.get_response(app)
        self.assertEqual(resp2.status_code, 200)

        self.assertIn('id="pDebugPanel-CSRFMultiScheme">', resp2.text)



##############################################################################
#
# Copyright (c) 2017 Jonathan Vanasco and Contributors
# Portions Copyright (c) 2008-2013 Agendaless Consulting and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################
import os

from setuptools import setup
from setuptools import find_packages

# store version in the init.py
import re

HERE = os.path.dirname(__file__)

long_description = (
    description
) = "Provides for creating independent csrf tokens for the http and https schemes"
with open(os.path.join(HERE, "README.rst")) as r_file:
    long_description = r_file.read()

with open(
    os.path.join(HERE, "src", "pyramid_csrf_multi_scheme", "__init__.py")
) as v_file:
    VERSION = re.compile(r'.*__VERSION__ = "(.*?)"', re.S).match(v_file.read()).group(1)


requires = [
    "pyramid>=1.10.4",
]
tests_require = [
    "pytest",
    "pyramid_debugtoolbar>=4.0.0",
]
testing_extras = tests_require + []

setup(
    name="pyramid_csrf_multi_scheme",
    version=VERSION,
    description=description,
    long_description=long_description,
    classifiers=[
        "Intended Audience :: Developers",
        "Framework :: Pyramid",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: Repoze Public License",
    ],
    keywords="web pyramid csrf",
    packages=find_packages(
        where="src",
    ),
    package_dir={"": "src"},
    include_package_data=True,
    author="Jonathan Vanasco",
    author_email="jonathan@findmeon.com",
    url="https://github.com/jvanasco/pyramid_csrf_multi_scheme",
    license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
    zip_safe=False,
    tests_require=tests_require,
    extras_require={
        "testing": testing_extras,
    },
    test_suite="tests",
)
